#!/usr/bin/env python

# The MIT License
#
# Copyright (c) 2007 Michael Stone <michael@ashsong.net>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
A module for first-order relations.

Constructors
~~~~~~~~~~~~

>>> Relation()
Relation([])

>>> Relation([])
Relation([])

>>> z = Row(('z', 0))
>>> print z
Row([('z', 0)])

>>> t = Row(('x', 1), ('y', 2))
>>> print t
Row([('x', 1), ('y', 2)])

>>> r = table(('x', 'y'), (1, 1), (2, 1), (3, 1))
>>> print r
Relation([Row([('x', 1), ('y', 1)]), Row([('x', 3), ('y', 1)]), Row([('x', 2), ('y', 1)])])

>>> r2 = column('z', 0)
>>> print r2
Relation([Row([('z', 0)])])




Coproducts   <--- is this correct?
~~~~~~~~~~

(a.k.a. "union")

>>> r + t
Relation([Row([('x', 1), ('y', 2)]), Row([('x', 1), ('y', 1)]), Row([('x', 3), ('y', 1)]), Row([('x', 2), ('y', 1)])])

>>> r + r
Relation([Row([('x', 1), ('y', 1)]), Row([('x', 3), ('y', 1)]), Row([('x', 2), ('y', 1)])])




Products     <---  is this correct?
~~~~~~~~

(a.k.a. "joins")

>>> r * z
Relation([Row([('x', 2), ('y', 1), ('z', 0)]), Row([('x', 1), ('y', 1), ('z', 0)]), Row([('x', 3), ('y', 1), ('z', 0)])])

>>> r * r2
Relation([Row([('x', 2), ('y', 1), ('z', 0)]), Row([('x', 1), ('y', 1), ('z', 0)]), Row([('x', 3), ('y', 1), ('z', 0)])])




Filtering
~~~~~~~~~

>>> (r * z) | (lambda a: a.x == 2)
Relation([Row([('x', 2), ('y', 1), ('z', 0)])])




Order
~~~~~

These are the obvious properties an ordering should have:

>>> column('z', 0) <= column('z', 0, 1)
True

>>> column('z', 0) >= column('z', 0, 1)
False

Arguably, it would be desirable for some of these ordering properties to hold:

>>> column('z', 0) <= table(('a', 'z'), (0, 0))
False

>>> column('z', 0) >= table(('a', 'z'), (0, 0))
False

>>> column('z', 1) <= table(('a', 'z'), (0, 0))
False

>>> column('z', 1) >= table(('a', 'z'), (0, 0))
False




Tabular Output
~~~~~~~~~~~~~~

>>> print sorted(set(r % ('x',)))
[(1,), (2,), (3,)]

>>> print sorted(set(r % ('y',)))
[(1,)]

>>> print sorted(r % ('y', 'x'))
[(1, 1), (1, 2), (1, 3)]




Modifications
~~~~~~~~~~~~~

>>> def twiddle(a):
...     del a.y
...     a.id = chr(ord('a') + a.x * a['x'] + 1)
...     if a.id not in 'hijklmnopqrstuvwxyz':
...         yield a
...         a.id = a.id.upper()
...         a.x *= 3
...         yield a
>>> sorted(r.transform(twiddle) % ('id', 'x'))
[('C', 3), ('F', 6), ('c', 1), ('f', 2)]
"""

from itertools import *

__all__ = ['table', 'column', 'Row', 'Relation']

def table(cols, *vals):
    return Relation(Row(*izip(cols, v)) for v in vals)

def column(name, *vals):
    return Relation(Row((name, val)) for val in vals)

class Row(frozenset):
    """Rows are intended to be (mathematica) functions on finite domains."""

    def __new__(cls, *args):
        return super(Row, cls).__new__(cls, args)

    def _filter(self, pred):
        return Row(*ifilter(pred, self))

    def _transform(self, fn):
        ret = fn(MutableRow(self))
        if ret is not None:
            return (Row(*b.iteritems()) for b in ret)

    def _compose(self, rhs, *cols):
        if len(cols) == 0:
            return self.union(rhs)
        colset = frozenset(cols)
        if colset.issubset(self.intersection(rhs)):
            return Row(a for a in self.union(rhs) if a[0] not in colset)
        else:
            return None

    def __getattr__(self, name):
        return dict(self)[name]

    def __getitem__(self, name):
        return dict(self)[name]

    def __add__(self, rhs):
        return Relation(self) + rhs

    def __mul__(self, rhs):
        return Relation(self) * rhs

    def __repr__(self):
        return "Row(%r)" % sorted(self)





class MutableRow(dict):
    """Rows must be immutable because they must be hashable. However, it is
    sometimes convenient (e.g. when transforming a relation) to be able to
    mutate rows. This facade is used by the Relation class to provide the
    desired interface.
    """
    def __init__(self, row=None):
        if row is not None:
            for k, v in dict(row).iteritems():
                self[k] = v

    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, val):
        self[name] = val

    def __delattr__(self, name):
        del self[name]






class Relation(object):
    """An immutable first-order relation.

Operations:

concatenate, r + r, r + t,
    union the rows of several relations

remove,      r - r, r - t,
    remove a set of rows from a relation

compose,     r * r, r * t,
    join two relations along a (possibly empty) set of columns

project,     r / (c1, c2),
    remove the given columns from a relation

filter,      r | (lambda a: a.age > 5),
    select rows to keep

transform,   r.transform(fn)
    r -> chain(*map(fn, r))

format,      r % (c1, c2),
    tabulate r for easy iteration
    """

    def __init__(self, iterable=None):
        if iterable:
            if isinstance(iterable, Row):
                self._data = set((iterable,))
                self._iter = None
            else:
                self._data = None
                self._iter = iterable
        else:
            self._data = set()
            self._iter = None

    def _force(self):
        if self._iter:
            self._data = set(self._iter)
            self._iter = None
        return self

    def __repr__(self):
        return "Relation(%r)" % sorted(self._force()._data)

    def __iter__(self):
        return iter(self._force()._data)

    def copy(self):
        return Relation(self._force()._data.copy())

    def project(self, *cols):
        self._iter = (tup._filter(lambda t: t[0] in cols) for tup in iter(self))
        self._data = None
        return self

    def compose(self, other, *cols):
        self._iter = ifilter(bool, (a._compose(b, *cols) for a in iter(self) for b in iter(other)))
        self._data = None
        return self

    def filter(self, pred):
        self._iter = ifilter(lambda t: pred(MutableRow(t)), iter(self))
        self._data = None
        return self

    def transform(self, fn):
        self._iter = chain(*(row._transform(fn) for row in iter(self)))
        self._data = None
        return self

    def concatenate(self, rhs):
        self._iter = chain(iter(self), iter(rhs))
        self._data = None
        return self

    def remove(self, rhs):
        self._force()
        self._data -= Relation(rhs)._force()._data
        return self

    def __contains__(self, item):
        # Because we're implementing first-order relations, and because we're
        # trying to transparently include bare sets of pairs as rows, we
        # choose to define containment in terms of our partial order on relations
        return Relation(item) <= self

    def __le__(self, rhs):
        return self._force()._data <= Relation(rhs)._force()._data

    def __lt__(self, rhs):
        return self._force()._data < Relation(rhs)._force()._data

    def __ge__(self, rhs):
        return self._force()._data >= Relation(rhs)._force()._data

    def __gt__(self, rhs):
        return self._force()._data > Relation(rhs)._force()._data

    def __eq__(self, rhs):
        return self._force()._data == Relation(rhs)._force()._data

    def __ne__(self, rhs):
        return self._force()._data != Relation(rhs)._force()._data

    def __add__(self, rhs):
        return self.copy().concatenate(Relation(rhs))

    def __concat__(self, rhs):
        return self.concatenate(Relation(rhs))

    def __sub__(self, rhs):
        return self.copy().remove(Relation(rhs))

    def __mul__(self, rhs):
        return self.copy().compose(Relation(rhs))

    def __mod__(self, names):
        for row in self.copy().project(*names):
            yield tuple(row[n] for n in names)

    def __or__(self, pred):
        return self.copy().filter(pred)


if __name__ == '__main__':
    import doctest
    doctest.testmod()
